const build_options = @import("build_options");
const std = @import("std");
const ascii = std.ascii;
const fmt = std.fmt;
const fs = std.fs;
const heap = std.heap;
const io = std.io;
const log = std.log.scoped(.stacktile);
const mem = std.mem;
const meta = std.meta;
const os = std.os;
const debug = std.debug;
const math = std.math;

const wayland = @import("wayland");
const wl = wayland.client.wl;
const river = wayland.client.river;
const zriver = wayland.client.zriver;

const ini = @import("ini.zig");

// Welcome to the "so how does this work again?"-comment. I wrote this because I
// don't work on this often enough to remember how (and why) I made this work.
// Enjoy.
//
// First, we parse the config file. Inside we find [layout:<name>] and
// [output:<name>] sections. For the latter we create an OutputConfig, which
// stores an ident_str so we can later match it to an Output as well as a list
// of pointers to Layout's. For the former we create a Layout.
//
// A Layout holds a MetaLayout, a List of SubLayout's and a default LayoutConfig.
// SubLayout's hold a configurable maximum amount of views, except for the last
// one, which has no view limit (by necesseity). The MetaLayout controls how the
// sublayouts are arranged, according to the LayoutConfig.Gravity.
//
// As expected, every Wayland output gets an Output. Each Output holds a
// LayoutConfig, for when the user overrides the active values via a user
// command. It also holds a hash map of LayoutConfig's, for when per tag layouts
// are active.
//
// Output's are matched OutputConfig's. Only if we habe a mathing pair do we
// bind the layout interface for the output and therefore handle layout demands.
// If we want to arrange an output, we will check the Output's OutputConfig's
// Layout list. Views are assigned locations and dimensions according to the
// configured MetaLayout and SubLayout's, only going to the next SubLayout when
// the current ones view limit is reached. The per-output / per-tag
// modfied LayoutConfig is used if it exists, otherwise fall back to the layouts
// default one.

pub const Gravity = enum { left, right, top, bottom };

const Box = struct {
    x: u31,
    y: u31,
    width: u32,
    height: u32,

    fn addOuterGap(self: *Box, gap: u31) Box {
        return .{
            .x = self.x +| gap,
            .y = self.y +| gap,
            .width = self.width -| (2 *| gap),
            .height = self.height -| (2 *| gap),
        };
    }

    /// Splits the box into two. Old box will be shrunk, new box according to
    /// gravity and divide will be returned.
    fn split(self: *Box, gravity: Gravity, ratio: f16, gap: u31) Box {
        debug.assert(ratio < 1.0 and ratio > 0.0);
        switch (gravity) {
            .left, .right => {
                var ret = Box{
                    .width = @as(u31, @intFromFloat((@as(f16, @floatFromInt(self.width)) * ratio) - (@as(f16, @floatFromInt(gap)) / @as(f16, 2)))),
                    .height = self.height,
                    .x = undefined,
                    .y = self.y,
                };
                self.width -= ret.width + gap;
                ret.x = if (gravity == .left) self.x else self.x + @as(u31, @intCast(self.width)) + gap;
                if (gravity == .left) self.x += @as(u31, @intCast(ret.width)) + gap;
                return ret;
            },
            .top, .bottom => {
                var ret = Box{
                    .width = self.width,
                    .height = @as(u31, @intFromFloat(@as(f16, (@as(f16, @floatFromInt(self.height)) * ratio) - (@as(f16, @floatFromInt(gap)) / @as(f16, 2))))),
                    .x = self.x,
                    .y = undefined,
                };
                self.height -= ret.height + gap;
                ret.y = if (gravity == .top) self.y else self.y + @as(u31, @intCast(self.height)) + gap;
                if (gravity == .top) self.y += @as(u31, @intCast(ret.height)) + gap;
                return ret;
            },
        }
    }

    /// Split box into three parts. Center Box size depends on ratio.
    fn triSplit(self: *Box, gravity: Gravity, ratio: f16, gap: u31) struct { center: Box, a: Box, b: Box } {
        switch (gravity) {
            .left, .right => {
                const center_width = @as(u32, @intFromFloat(ratio * @as(f16, @floatFromInt(self.width))));
                const other_width = @divFloor(self.width - center_width, 2);
                return .{
                    .center = .{
                        .x = self.x + @as(u31, @intCast(other_width)) + @divFloor(gap, 2),
                        .y = self.y,
                        .width = center_width - gap,
                        .height = self.height,
                    },
                    .a = .{
                        .x = self.x + (if (gravity == .right) 0 else @as(u31, @intCast(other_width + center_width + @divFloor(gap, 2)))),
                        .y = self.y,
                        .width = other_width - @divFloor(gap, 2),
                        .height = self.height,
                    },
                    .b = .{
                        .x = self.x + (if (gravity == .left) 0 else @as(u31, @intCast(other_width + center_width + @divFloor(gap, 2)))),
                        .y = self.y,
                        .width = other_width - @divFloor(gap, 2),
                        .height = self.height,
                    },
                };
            },
            .top, .bottom => {
                const center_height = @as(u32, @intFromFloat(ratio * @as(f16, @floatFromInt(self.height))));
                const other_height = @divFloor(self.height - center_height, 2);
                return .{
                    .center = .{
                        .x = self.x,
                        .y = self.y + @as(u31, @intCast(other_height)) + @divFloor(gap, 2),
                        .width = self.width,
                        .height = center_height - gap,
                    },
                    .a = .{
                        .x = self.x,
                        .y = self.y + (if (gravity == .bottom) 0 else @as(u31, @intCast(other_height + center_height + @divFloor(gap, 2)))),
                        .width = self.width,
                        .height = other_height - @divFloor(gap, 2),
                    },
                    .b = .{
                        .x = self.x,
                        .y = self.y + (if (gravity == .top) 0 else @as(u31, @intCast(other_height + center_height + @divFloor(gap, 2)))),
                        .width = self.width,
                        .height = other_height - @divFloor(gap, 2),
                    },
                };
            },
        }
    }

    pub const LinearMultiSplitIterator = struct {
        box: Box,
        gravity: Gravity,
        gap: u31,
        amount: usize,
        i: usize = 0,

        pub fn next(self: *LinearMultiSplitIterator) ?Box {
            if (self.i == self.amount) return null;
            defer self.i += 1;
            switch (self.gravity) {
                .left => return Box{
                    .x = @as(u31, @intCast(self.box.x +| (self.i *| (self.box.width + self.gap)))),
                    .y = self.box.y,
                    .width = self.box.width,
                    .height = self.box.height,
                },
                .right => return Box{
                    .x = @as(u31, @intCast(self.box.x +| ((self.amount -| self.i -| 1) *| (self.box.width + self.gap)))),
                    .y = self.box.y,
                    .width = self.box.width,
                    .height = self.box.height,
                },
                .top => return Box{
                    .x = self.box.x,
                    .y = @as(u31, @intCast(self.box.y +| (self.i *| (self.box.height + self.gap)))),
                    .width = self.box.width,
                    .height = self.box.height,
                },
                .bottom => return Box{
                    .x = self.box.x,
                    .y = @as(u31, @intCast(self.box.y +| ((self.amount -| self.i -| 1) *| (self.box.height + self.gap)))),
                    .width = self.box.width,
                    .height = self.box.height,
                },
            }
        }
    };

    fn linearMultiSplitIterator(self: *const Box, gravity: Gravity, gap: u31, amount: usize) LinearMultiSplitIterator {
        return .{
            .gravity = gravity,
            .gap = gap,
            .amount = amount,
            .box = .{
                .x = self.x,
                .y = self.y,
                .width = switch (gravity) {
                    .left, .right => @as(u32, @intCast(@divFloor(self.width -| ((amount -| 1) *| gap), amount))),
                    .top, .bottom => self.width,
                },
                .height = switch (gravity) {
                    .left, .right => self.height,
                    .top, .bottom => @as(u32, @intCast(@divFloor(self.height -| ((amount -| 1) *| gap), amount))),
                },
            },
        };
    }
};

test "Box" {
    const testing = std.testing;
    const amount = 4;
    var box = Box{
        .x = 100,
        .y = 200,
        .width = 5000,
        .height = 6000,
    };
    var it = box.linearMultiSplitIterator(.left, 20, amount);
    var i: usize = 0;
    while (it.next()) |_| : (i += 1) {}
    try testing.expect(i == amount);
}

const OverrideConfig = enum { @"per-output", @"per-tag", @"per-dominant-tag" };

const Context = struct {
    loop: bool = true,
    layout_manager: ?*river.LayoutManagerV3 = null,
    gpa: heap.GeneralPurposeAllocator(.{}) = .{},
    config_arena: heap.ArenaAllocator = undefined,
    outputs: std.TailQueue(Output) = .{},
    output_configs: std.ArrayListUnmanaged(OutputConfig) = .{},
    layouts: std.ArrayListUnmanaged(Layout) = .{},
    override_config: OverrideConfig = .@"per-dominant-tag",

    fn newOutput(self: *Context, registry: *wl.Registry, global_name: u32) !void {
        log.debug("binding output {}", .{global_name});
        const wl_output = try registry.bind(global_name, wl.Output, 4);
        errdefer wl_output.release();

        const alloc = self.gpa.allocator();
        const node = try alloc.create(std.TailQueue(Output).Node);
        errdefer alloc.destroy(node);

        try node.data.init(wl_output, global_name);
        self.outputs.append(node);
    }

    fn newLayout(self: *Context, path: []const u8, line: usize, name: []const u8) !*Layout {
        if (name.len == 0) {
            log.err("{s}:{}: missing layout name", .{ path, line });
            return error.BadConfig;
        }

        for (self.layouts.items) |l| {
            if (mem.eql(u8, l.name, name)) {
                log.err("{s}:{}: duplicate layout name: \"{s}\"", .{ path, line, name });
                return error.BadConfig;
            }
        }

        log.debug("adding layout: \"{s}\"", .{name});
        const alloc = self.config_arena.allocator();
        const ret = try context.layouts.addOne(alloc);
        try ret.sublayouts.ensureTotalCapacity(alloc, 10);
        ret.* = .{ .name = try alloc.dupeZ(u8, name) };
        return ret;
    }

    fn newOutputConfig(self: *Context, path: []const u8, line: usize, ident_str: []const u8) !*OutputConfig {
        if (ident_str.len == 0) {
            log.err("{s}:{}: missing output identifier", .{ path, line });
            return error.BadConfig;
        }

        for (self.output_configs.items) |oc| {
            if (mem.eql(u8, oc.ident_str, ident_str)) {
                log.err("{s}:{}: duplicate output configuration: \"{s}\"", .{ path, line, ident_str });
                return error.BadConfig;
            }
        }

        log.debug("adding output config: \"{s}\"", .{ident_str});
        const alloc = self.config_arena.allocator();
        const ret = try context.output_configs.addOne(alloc);
        try ret.layouts.ensureTotalCapacity(alloc, 5);
        ret.* = .{ .ident_str = try alloc.dupe(u8, ident_str) };
        return ret;
    }

    fn handleConfigAssign(self: *Context, path: []const u8, line: usize, variable: []const u8, value: []const u8) !void {
        if (mem.eql(u8, variable, "runtime-changes")) {
            self.*.override_config = meta.stringToEnum(OverrideConfig, value) orelse {
                log.err("{s}:{}: unknown value for setting \"{s}\": \"{s}\".", .{ path, line, variable, value });
                return error.BadConfig;
            };
        } else {
            log.err("{s}:{}: unknown setting: \"{s}\"", .{ path, line, variable });
            return error.BadConfig;
        }
    }
};

const LayoutConfig = struct {
    gravity: ?Gravity = null,
    outer_padding: ?u31 = null,
    main_padding: ?u31 = null,
    metalayout_padding: ?u31 = null,
    sublayout_padding: ?u31 = null,
    main_ratio: ?f16 = null,
    main_limit: ?usize = null,

    /// Parse assignments. Needs the active layout to know what the initial
    /// values for increments / decrements are.
    fn assign(self: *LayoutConfig, layout: *Layout, variable: []const u8, value: []const u8, operation: []const u8) !void {
        const lc = layout.*.layout_config;
        if (mem.eql(u8, variable, "gravity")) {
            if (!mem.eql(u8, operation, "=")) return error.UnknownOperation;
            self.*.gravity = meta.stringToEnum(Gravity, value) orelse return error.UnknownValue;
        } else if (mem.eql(u8, variable, "outer-padding")) {
            self.*.outer_padding = try intAssign(u31, self.*.outer_padding, lc.outer_padding.?, value, operation);
        } else if (mem.eql(u8, variable, "main-padding")) {
            self.*.main_padding = try intAssign(u31, self.*.main_padding, lc.main_padding.?, value, operation);
        } else if (mem.eql(u8, variable, "main-limit")) {
            const primary_sl = layout.sublayouts.items[0];
            self.*.main_limit = @max(1, try intAssign(usize, self.*.main_limit, primary_sl.getViewAmountLimit(), value, operation));
        } else if (mem.eql(u8, variable, "metalayout-padding")) {
            self.*.metalayout_padding = try intAssign(u31, self.*.metalayout_padding, lc.metalayout_padding.?, value, operation);
        } else if (mem.eql(u8, variable, "sublayout-padding")) {
            self.*.sublayout_padding = try intAssign(u31, self.*.sublayout_padding, lc.sublayout_padding.?, value, operation);
        } else if (mem.eql(u8, variable, "all-padding")) {
            self.*.sublayout_padding = try intAssign(u31, self.*.sublayout_padding, lc.sublayout_padding.?, value, operation);
            self.*.metalayout_padding = try intAssign(u31, self.*.metalayout_padding, lc.metalayout_padding.?, value, operation);
            self.*.main_padding = try intAssign(u31, self.*.main_padding, lc.main_padding.?, value, operation);
            self.*.outer_padding = try intAssign(u31, self.*.outer_padding, lc.outer_padding.?, value, operation);
        } else if (mem.eql(u8, variable, "main-ratio")) {
            self.*.main_ratio = try floatAssign(f16, self.*.main_ratio, lc.main_ratio.?, value, operation, 0.05, 0.95);
        } else {
            return error.UnknownVariable;
        }
    }

    fn intAssign(comptime T: type, _old: ?T, layout_value: T, _value: []const u8, operation: []const u8) !T {
        const old = if (_old) |o| o else layout_value;
        if (mem.eql(u8, operation, "=")) {
            // Find match and then return next value.
            var it = mem.split(u8, _value, ":");
            while (it.next()) |token| {
                const t = fmt.parseInt(T, token, 10) catch return error.BadInt;
                if (t != old) continue;

                // If there is no next value, wrap around to first one.
                const next_token = it.next() orelse blk: {
                    var iit = mem.split(u8, _value, ":");
                    break :blk iit.next().?;
                };
                return fmt.parseInt(T, next_token, 10) catch return error.BadInt;
            }

            // Fall back to first value in list if no matches are found.
            it = mem.split(u8, _value, ":");
            const token = it.next() orelse _value;
            return fmt.parseInt(T, token, 10) catch return error.BadInt;
        } else if (mem.eql(u8, operation, "+=")) {
            const value = fmt.parseInt(T, _value, 10) catch return error.BadInt;
            return old +| value;
        } else if (mem.eql(u8, operation, "-=")) {
            const value = fmt.parseInt(T, _value, 10) catch return error.BadInt;
            return old -| value;
        } else {
            return error.UnknownOperation;
        }
    }

    fn floatAssign(comptime T: type, _old: ?T, layout_value: T, _value: []const u8, operation: []const u8, lower: T, upper: T) !T {
        const value = fmt.parseFloat(T, _value) catch return error.BadFloat;
        const old = if (_old) |o| o else layout_value;
        if (mem.eql(u8, operation, "=")) {
            return math.clamp(value, lower, upper);
        } else if (mem.eql(u8, operation, "+=")) {
            return math.clamp(old + value, lower, upper);
        } else if (mem.eql(u8, operation, "-=")) {
            return math.clamp(old - value, lower, upper);
        } else {
            return error.UnknownOperation;
        }
    }
};

const Layout = struct {
    pub const MetaLayout = enum {
        linear,
        orthogonal,
        @"linear-center",
        @"orthogonal-center",
    };

    const SubLayout = union(enum) {
        full: usize,
        stack: usize,
        rows: usize,
        columns: usize,
        orthogonal: usize,
        parallel: usize,

        fn getViewAmountLimit(self: SubLayout) usize {
            // TODO this is probably elegantly solvable with some zig magic
            return switch (self) {
                .full => |i| i,
                .stack => |i| i,
                .rows => |i| i,
                .columns => |i| i,
                .orthogonal => |i| i,
                .parallel => |i| i,
            };
        }
    };

    name: [:0]const u8,
    display_name: ?[:0]const u8 = null,
    metalayout: MetaLayout = .linear, // TODO should this be part of LayoutConfig?
    sublayouts: std.ArrayListUnmanaged(SubLayout) = .{},
    layout_config: LayoutConfig = .{
        .outer_padding = 0,
        .main_padding = 0,
        .main_ratio = 0.6,
        .metalayout_padding = 0,
        .sublayout_padding = 0,
        .gravity = .left,
    },

    rule_tags: ?u32 = null,

    fn handleConfigAssign(self: *Layout, path: []const u8, line: usize, variable: []const u8, value: []const u8) !void {
        if (mem.eql(u8, variable, "display-name")) {
            if (self.display_name != null) {
                log.err("{s}:{}: 'display-name' may not be specified more than once.", .{ path, line });
                return error.BadConfig;
            }
            const alloc = context.config_arena.allocator();
            self.display_name = try alloc.dupeZ(u8, value);
        } else if (mem.eql(u8, variable, "metalayout")) {
            self.metalayout = meta.stringToEnum(MetaLayout, value) orelse {
                log.err("{s}:{}: unknown metalayout: \"{s}\".", .{ path, line, value });
                return error.BadConfig;
            };
        } else if (mem.eql(u8, variable, "sublayouts")) {
            try self.parseSublayouts(path, line, value);
        } else if (mem.eql(u8, variable, "tags")) {
            if (mem.eql(u8, value, "all")) {
                self.rule_tags = math.maxInt(u32);
            } else {
                self.rule_tags = fmt.parseInt(u32, value, 10) catch {
                    log.err("{s}:{}: bad tag: \"{s}\"", .{ path, line, value });
                    return error.BadConfig;
                };
            }
        } else {
            self.layout_config.assign(self, variable, value, "=") catch |err| {
                switch (err) {
                    error.UnknownVariable => log.err("{s}:{}: unknown setting: \"{s}\".", .{ path, line, variable }),
                    error.UnknownValue => log.err("{s}:{}: unknown value for setting \"{s}\": \"{s}\".", .{ path, line, variable, value }),
                    error.BadInt => log.err("{s}:{}: bad integer: \"{s}\"", .{ path, line, value }),
                    error.BadFloat => log.err("{s}:{}: bad floating point number: \"{s}\"", .{ path, line, value }),
                    error.UnknownOperation => unreachable,
                }
                return error.BadConfig;
            };
        }
    }

    fn parseSublayouts(self: *Layout, path: []const u8, line: usize, value: []const u8) !void {
        const alloc = context.config_arena.allocator();
        var it = mem.tokenize(u8, value, &ascii.whitespace);
        while (it.next()) |token| {
            var iit = mem.split(u8, token, ":");
            const tag = iit.next().?;
            const amount: usize = blk: {
                if (iit.next()) |a| {
                    const b = try parseInt(usize, path, line, a);
                    if (b == 0) {
                        log.err("{s}:{}: sublayout view amount restriction must be greater than 0.", .{ path, line });
                        return error.BadConfig;
                    }
                    break :blk b;
                }
                break :blk 0;
            };

            // TODO this needs some zig magic
            if (mem.eql(u8, tag, "full")) {
                try self.sublayouts.append(alloc, .{ .full = amount });
            } else if (mem.eql(u8, tag, "stack")) {
                try self.sublayouts.append(alloc, .{ .stack = amount });
            } else if (mem.eql(u8, tag, "columns")) {
                try self.sublayouts.append(alloc, .{ .columns = amount });
            } else if (mem.eql(u8, tag, "rows")) {
                try self.sublayouts.append(alloc, .{ .rows = amount });
            } else if (mem.eql(u8, tag, "orthogonal")) {
                try self.sublayouts.append(alloc, .{ .orthogonal = amount });
            } else if (mem.eql(u8, tag, "parallel")) {
                try self.sublayouts.append(alloc, .{ .parallel = amount });
            } else {
                log.err("{s}:{}: unknown sublayout: \"{s}\"", .{ path, line, tag });
                return error.BadConfig;
            }
        }

        // Check if sublayouts make sense.
        var unlimited: bool = false;
        for (self.sublayouts.items) |sl| {
            if (unlimited) {
                log.err("{s}:{}: there may not be another sublayout after an unlimited sublayout", .{ path, line });
                return error.BadConfig;
            }
            if (sl.getViewAmountLimit() == 0) unlimited = true;
        }
        if (!unlimited) {
            log.err("{s}:{}: last sublayout is not allowed to have a view amount limit.", .{ path, line });
            return error.BadConfig;
        }
    }

    fn parseInt(comptime T: type, path: []const u8, line: usize, value: []const u8) !T {
        return fmt.parseInt(T, value, 10) catch |err| {
            log.err("{s}:{}: invalid integer: \"{s}\": {s}.", .{ path, line, value, @errorName(err) });
            return error.BadConfig;
        };
    }

    fn parseFloat(comptime T: type, path: []const u8, line: usize, value: []const u8) !T {
        return fmt.parseFloat(T, value) catch |err| {
            log.err("{s}:{}: invalid floating point number: \"{s}\": {s}.", .{ path, line, value, @errorName(err) });
            return error.BadConfig;
        };
    }

    pub fn getDisplayName(self: Layout) [:0]const u8 {
        if (self.display_name) |n| return n;
        return self.name;
    }
};

const OutputConfig = struct {
    /// Holds either a potential Layout name or a pointer to a Layout.
    /// Output configs and layouts can be defined in the config file in an
    /// arbitrary order. So first we store the desired layout name, then after
    /// we are done parsing the config file we can try and match it with an
    /// actual Layout.
    const MaybeLayout = union(enum) {
        name: []const u8,
        layout: *Layout,
    };

    ident_str: []const u8,
    layouts: std.ArrayListUnmanaged(MaybeLayout) = .{},

    fn handleConfigAssign(self: *OutputConfig, path: []const u8, line: usize, variable: []const u8, value: []const u8) !void {
        if (mem.eql(u8, variable, "layout") or mem.eql(u8, variable, "layouts")) {
            const alloc = context.config_arena.allocator();
            var it = mem.tokenize(u8, value, &ascii.whitespace);
            while (it.next()) |token| {
                const ll = try self.layouts.addOne(alloc);
                ll.* = .{ .name = try alloc.dupe(u8, token) };
            }
        } else {
            log.err("{s}:{}: unknown output config variable: \"{s}\".", .{ path, line, variable });
            return error.BadConfig;
        }
    }

    fn getLayouts(self: *const OutputConfig) !void {
        if (self.layouts.items.len == 0) {
            log.err("output configuration for \"{s}\" does not contain any layouts", .{self.ident_str});
            return error.BadConfig;
        }

        for (self.layouts.items) |*l| {
            debug.assert(l.* == .name);
            for (context.layouts.items) |*L| {
                if (mem.eql(u8, L.*.name, l.*.name)) {
                    l.* = .{ .layout = L };
                    break;
                }
            } else {
                log.err("layout \"{s}\" set for output \"{s}\" has not been defined", .{ l.*.name, self.ident_str });
                return error.BadConfig;
            }
        }
    }

    /// Return active Layout, based on rules. Should no layout fit, use last one.
    pub fn getActiveLayout(self: *OutputConfig, active_tags: u32) *Layout {
        for (self.layouts.items) |ml| {
            const layout = ml.layout;

            if (layout.rule_tags) |rule_tags| {
                if (rule_tags == active_tags) return layout;
            }
        }

        return self.layouts.items[self.layouts.items.len - 1].layout;
    }
};

const Output = struct {
    global_name: u32,
    wl_output: *wl.Output,
    name: ?[]const u8 = null,
    make_model: ?[]const u8 = null,
    description: ?[]const u8 = null,
    output_config: ?*OutputConfig = null,
    layout: ?*river.LayoutV3 = null,
    configured: bool = false,

    // The tags for the next user command. The protocol guarantees this to be
    // send directly before every user command.
    next_user_command_tags: u32 = undefined,

    // TODO move to OutputConfig if we ever want this to survive output hotplugging.
    layout_config_overrides: std.AutoHashMapUnmanaged(u32, LayoutConfig) = .{},

    fn init(self: *Output, wl_output: *wl.Output, global_name: u32) !void {
        self.* = .{
            .wl_output = wl_output,
            .global_name = global_name,
        };
        wl_output.setListener(*Output, outputListener, self);
    }

    fn deinit(self: *Output) void {
        const alloc = context.gpa.allocator();
        if (self.layout) |l| l.destroy();
        if (self.name) |n| alloc.free(n);
        if (self.make_model) |m| alloc.free(m);
        if (self.description) |d| alloc.free(d);
        self.wl_output.release();
    }

    fn outputListener(_: *wl.Output, event: wl.Output.Event, self: *Output) void {
        const alloc = context.gpa.allocator();
        switch (event) {
            .geometry => |ev| {
                log.debug("output {} advertised make and model: \"{s} {s}\"", .{ self.global_name, ev.make, ev.model });
                self.make_model = fmt.allocPrint(alloc, "{s} {s}", .{ ev.make, ev.model }) catch |err| {
                    log.err("failed to allocate memory: {s}", .{@errorName(err)});
                    return;
                };
            },
            .name => |ev| {
                log.debug("output {} advertised name: \"{s}\"", .{ self.global_name, ev.name });
                self.name = alloc.dupeZ(u8, mem.span(ev.name)) catch |err| {
                    log.err("failed to allocate memory: {s}", .{@errorName(err)});
                    return;
                };
            },
            .description => |ev| {
                log.debug("output {} advertised description: \"{s}\"", .{ self.global_name, ev.description });
                self.description = alloc.dupeZ(u8, mem.span(ev.description)) catch |err| {
                    log.err("failed to allocate memory: {s}", .{@errorName(err)});
                    return;
                };
            },
            .done => {
                // this means we don't respond to outputs updating their name
                // or description. That will probably never be an issue though,
                // right?
                if (self.output_config) |_| return;

                for (context.output_configs.items) |*oc| {
                    if (self.name) |name| {
                        if (mem.eql(u8, oc.*.ident_str, name)) {
                            self.*.output_config = oc;
                            break;
                        }
                    }
                    if (self.make_model) |make_model| {
                        if (mem.eql(u8, oc.*.ident_str, make_model)) {
                            self.*.output_config = oc;
                            break;
                        }
                    }
                    if (self.description) |description| {
                        if (mem.eql(u8, oc.*.ident_str, description)) {
                            self.*.output_config = oc;
                            break;
                        }
                    }
                }

                if (self.output_config) |oc| {
                    log.debug("output {} matched with output config: \"{s}\"", .{ self.global_name, oc.ident_str });
                    self.bindLayout() catch |err| {
                        log.err("failed to configure output {}: {s}", .{ self.global_name, @errorName(err) });
                    };
                }
            },
            else => {},
        }
    }

    fn bindLayout(self: *Output) !void {
        if (self.layout) |_| return;

        if (self.output_config == null) {
            // TODO add a way to define default config that matches all outputs.
            log.debug("output {} has no matching output config (yet), not binding layout (yet)", .{self.global_name});
            return;
        }

        if (context.layout_manager) |lm| {
            log.debug("output {} binding river_layout_v3", .{self.global_name});
            self.layout = try lm.getLayout(self.wl_output, "stacktile");
            self.layout.?.setListener(*Output, layoutListener, self);
        } else {
            log.debug(
                "can not get river_layout_v3 for output {}, river_layout_manager_v3 not yet bound",
                .{self.global_name},
            );
        }
    }

    fn layoutListener(layout: *river.LayoutV3, event: river.LayoutV3.Event, self: *Output) void {
        switch (event) {
            .namespace_in_use => {
                log.err("output {} layout namsepace already in use", .{self.global_name});
                self.layout.?.destroy();
                self.layout = null;
            },
            .user_command_tags => |ev| {
                log.debug("user_command_tags received: {}", .{ev.tags});
                self.next_user_command_tags = ev.tags;
            },
            .user_command => |ev| {
                // TODO cycle
                // TODO unset/reset

                // TODO use the last active one; If there is no last active one, use the first one.
                const active_layout = self.output_config.?.getActiveLayout(self.next_user_command_tags);

                // We support sending multiple commands at the same time, separated by a semicolon.
                var it = mem.tokenize(u8, mem.span(ev.command), ";");
                while (it.next()) |_cmd| {
                    var cmd_it = mem.tokenize(u8, _cmd, &ascii.whitespace);
                    const cmd = cmd_it.next() orelse continue;
                    if (mem.eql(u8, cmd, "exit")) {
                        if (cmd_it.next() != null) {
                            log.err("command \"exit\" does not support any arguments.", .{});
                        } else {
                            context.loop = false;
                        }
                    } else if (mem.eql(u8, cmd, "set")) {
                        const variable = cmd_it.next();
                        const operation = cmd_it.next();
                        const value = cmd_it.next();
                        if (variable == null or operation == null or value == null) {
                            log.err("too few arguments for user command \"set\"", .{});
                            continue;
                        }
                        const mlc = self.getMutableLayoutConfig(self.next_user_command_tags) catch |err| {
                            log.err("{s}", .{@errorName(err)});
                            continue;
                        };
                        mlc.assign(active_layout, variable.?, value.?, operation.?) catch |err| {
                            switch (err) {
                                error.UnknownVariable => log.err("unknown setting: \"{s}\".", .{variable.?}),
                                error.UnknownValue => log.err("unknown value for setting \"{s}\": \"{s}\".", .{ variable.?, value.? }),
                                error.UnknownOperation => log.err("unsupported operation for setting \"{s}\": \"{s}\".", .{ value.?, variable.? }),
                                error.BadInt => log.err("bad integer: \"{s}\"", .{value.?}),
                                error.BadFloat => log.err("bad floating point number: \"{s}\"", .{value.?}),
                            }
                            continue;
                        };
                    } else if (mem.eql(u8, cmd, "resize-main")) {
                        const direction = cmd_it.next();
                        const value = cmd_it.next();
                        if (direction == null or value == null) {
                            log.err("too few arguments for user command \"resize-main\"", .{});
                            continue;
                        }
                        self.directionalResize(active_layout, direction.?, value.?) catch |err| {
                            switch (err) {
                                error.BadDirection => log.err("unknown direction. \"{s}\"", .{direction.?}),
                                error.BadFloat => log.err("bad floating point number: \"{s}\"", .{value.?}),
                                else => unreachable,
                            }
                        };
                    } else {
                        log.err("unknown command: \"{s}\"", .{cmd});
                        continue;
                    }
                }
            },
            .layout_demand => |ev| {
                log.debug("output {} tag {} layout demand with {} view(s)", .{ self.global_name, ev.tags, ev.view_count });
                debug.assert(self.output_config != null);

                const active_layout = self.output_config.?.getActiveLayout(ev.tags);
                const lc = self.getActiveLayoutConfig(active_layout, ev.tags);

                const primary_sl = active_layout.sublayouts.items[0];
                const main_limit = blk: {
                    if (lc.main_limit) |l| break :blk l;
                    break :blk primary_sl.getViewAmountLimit();
                };

                // First let's find out how many zones our metalayout needs.
                const zones = blk: {
                    if (main_limit == 0 or main_limit >= ev.view_count) {
                        break :blk 1;
                    }

                    // We can be sure that there is more than one sublayout,
                    // since the last sublayout is known to have a limit of 0.
                    var zones: usize = 1;
                    var vc: usize = ev.view_count - main_limit;
                    for (active_layout.sublayouts.items[1..]) |sl| {
                        zones += 1;
                        const vl = sl.getViewAmountLimit();
                        if (vl == 0 or vl >= vc) break;
                        vc -= vl;
                    }
                    break :blk zones;
                };

                const box_bind = (Box{
                    .x = 0,
                    .y = 0,
                    .width = ev.usable_width,
                    .height = ev.usable_height,
                });
                var box = @constCast(&box_bind).addOuterGap(lc.outer_padding.?);

                if (zones == 1) {
                    self.doSubLayout(lc, ev.view_count, active_layout.sublayouts.items[0], box, ev.serial);
                } else if (active_layout.metalayout == .linear or
                    active_layout.metalayout == .orthogonal or
                    zones == 2)
                {
                    // Primary zone.
                    const primary_box = box.split(lc.gravity.?, lc.main_ratio.?, lc.main_padding.?);
                    self.doSubLayout(lc, main_limit, primary_sl, primary_box, ev.serial);

                    // Rest of the zones.
                    var view_count = ev.view_count - main_limit;
                    try self.doMetaLayoutLinear(active_layout, lc, box, zones - 1, 0, &view_count, ev.serial);
                } else {
                    const split = box.triSplit(lc.gravity.?, lc.main_ratio.?, lc.main_padding.?);
                    self.doSubLayout(lc, main_limit, primary_sl, split.center, ev.serial);

                    const a_zones = @divFloor(zones - 1, 2);
                    const b_zones = zones - 1 - a_zones;

                    var view_count = ev.view_count - main_limit;

                    try self.doMetaLayoutLinear(active_layout, lc, split.a, a_zones, 0, &view_count, ev.serial);
                    try self.doMetaLayoutLinear(active_layout, lc, split.b, b_zones, a_zones, &view_count, ev.serial);
                }

                layout.commit(active_layout.getDisplayName().ptr, ev.serial);
            },
        }
    }

    fn doMetaLayoutLinear(
        self: *Output,
        layout: *Layout,
        lc: LayoutConfig,
        box: Box,
        zones: usize,
        skip_zones: usize,
        view_count: *usize,
        serial: u32,
    ) !void {
        debug.assert(view_count.* > 0);
        // Local effective gravity, "calculated" from the gravity of the layout
        // config and the metalayout. Note that this is different from the
        // configured gravity, it is merely the gravity used for the linear
        // Box stack calculated here.
        const local_gravity = blk: { // TODO[zig]: should be two nested switches, but compiler doesn't like this yet...
            if (layout.metalayout == .linear or layout.metalayout == .@"linear-center") break :blk lc.gravity.?;
            if (layout.metalayout == .orthogonal or layout.metalayout == .@"orthogonal-center") {
                switch (lc.gravity.?) {
                    .top, .bottom => break :blk .left, // TODO secondary-gravity, or maybe "orientation" (in the mathematical sense)
                    .left, .right => break :blk .top,
                }
            }
            unreachable;
        };
        var it = box.linearMultiSplitIterator(local_gravity, lc.metalayout_padding.?, zones);
        var i: usize = 1; // We start at one, since the primary zone is handled elsewhere.
        while (it.next()) |bx| : (i += 1) {
            const sublayout = layout.sublayouts.items[skip_zones + i];
            const zone_view_count = blk: {
                const zone_limit = sublayout.getViewAmountLimit();
                if (zone_limit == 0) break :blk view_count.*;
                break :blk @min(view_count.*, zone_limit);
            };
            self.doSubLayout(lc, zone_view_count, sublayout, bx, serial);
            view_count.* -|= zone_view_count;
        }
        debug.assert(i == zones + 1);
    }

    fn doSubLayout(
        self: *Output,
        lc: LayoutConfig,
        view_count: usize,
        sublayout: Layout.SubLayout,
        box: Box,
        serial: u32,
    ) void {
        if (view_count == 1) {
            self.layout.?.pushViewDimensions(
                @as(i32, box.x),
                @as(i32, box.y),
                box.width,
                box.height,
                serial,
            );
            return;
        }

        switch (sublayout) {
            .full => {
                var vc = view_count;
                while (vc > 0) : (vc -= 1) {
                    self.layout.?.pushViewDimensions(
                        @as(i32, box.x),
                        @as(i32, box.y),
                        box.width,
                        box.height,
                        serial,
                    );
                }
            },
            .stack => {
                const width = @as(u31, @intFromFloat((0.95 * @as(f16, @floatFromInt(box.width)))));
                const height = @as(u31, @intFromFloat((0.95 * @as(f16, @floatFromInt(box.height)))));
                const x_offset = @as(u31, @intFromFloat((0.05 * @as(f16, @floatFromInt(box.width)) / @as(f16, @floatFromInt(view_count - 1)))));
                const y_offset = @as(u31, @intFromFloat((0.05 * @as(f16, @floatFromInt(box.height)) / @as(f16, @floatFromInt(view_count - 1)))));
                var i: usize = 0;
                while (i < view_count) : (i += 1) {
                    self.layout.?.pushViewDimensions(
                        @as(i32, @intCast(box.x + (i * x_offset))),
                        @as(i32, @intCast(box.y + (i * y_offset))),
                        width,
                        height,
                        serial,
                    );
                }
            },
            .rows, .columns, .orthogonal, .parallel => {
                // TODO[zig]: should be two nested switches, but compiler doesn't like that yet...
                const local_gravity = blk: {
                    if (sublayout == .rows) {
                        if (lc.gravity.? == .top or lc.gravity.? == .bottom) {
                            break :blk lc.gravity.?;
                        } else {
                            break :blk .top; // TODO secondary gravity
                        }
                    } else if (sublayout == .columns) {
                        if (lc.gravity.? == .top or lc.gravity.? == .bottom) {
                            break :blk .left; // TODO secondary gravity
                        } else {
                            break :blk lc.gravity.?;
                        }
                    } else if (sublayout == .orthogonal) {
                        if (lc.gravity.? == .top or lc.gravity.? == .bottom) {
                            break :blk .left; // TODO secondary gravity
                        } else {
                            break :blk .top; // TODO secondary gravity
                        }
                    } else if (sublayout == .parallel) {
                        break :blk lc.gravity.?;
                    } else {
                        unreachable;
                    }
                };
                var it = box.linearMultiSplitIterator(local_gravity, lc.sublayout_padding.?, view_count);
                while (it.next()) |b| {
                    self.layout.?.pushViewDimensions(b.x, b.y, b.width, b.height, serial);
                }
            },
        }
    }

    fn directionalResize(self: *Output, layout: *Layout, _direction: []const u8, value: []const u8) !void {
        const lc = self.getActiveLayoutConfig(layout, self.next_user_command_tags);
        const direction = meta.stringToEnum(Gravity, _direction) orelse return error.BadDirection;
        const operation = blk: {
            switch (lc.gravity.?) {
                .top => switch (direction) {
                    .top => break :blk "-=",
                    .bottom => break :blk "+=",
                    .left, .right => return,
                },
                .bottom => switch (direction) {
                    .top => break :blk "+=",
                    .bottom => break :blk "-=",
                    .left, .right => return,
                },
                .left => switch (direction) {
                    .left => break :blk "-=",
                    .right => break :blk "+=",
                    .top, .bottom => return,
                },
                .right => switch (direction) {
                    .left => break :blk "+=",
                    .right => break :blk "-=",
                    .top, .bottom => return,
                },
            }
        };
        const mlc = self.getMutableLayoutConfig(self.next_user_command_tags) catch |err| {
            log.err("{}", .{err});
            return;
        };
        try mlc.assign(layout, "main-ratio", value, operation);
    }

    /// Gather all active layout variables and combine them into a single
    /// LayoutConfig object. Note that this object is not mutable.
    fn getActiveLayoutConfig(self: *Output, layout: *Layout, tags: u32) LayoutConfig {
        var ret: LayoutConfig = layout.*.layout_config;
        if (self.layout_config_overrides.get(getLayoutConfigKey(tags))) |override| {
            const t = @typeInfo(LayoutConfig);
            inline for (t.Struct.fields) |field| {
                if (@field(override, field.name)) |val| @field(ret, field.name) = val;
            }
        }
        return ret;
    }

    fn getMutableLayoutConfig(self: *Output, tags: u32) !*LayoutConfig {
        const alloc = context.config_arena.allocator();
        const key = getLayoutConfigKey(tags);
        const ret = try self.layout_config_overrides.getOrPut(alloc, key);
        if (!ret.found_existing) {
            log.debug("creating layout config override for output {} with key {}", .{ self.global_name, key });
            ret.value_ptr.* = .{};
        }
        return ret.value_ptr;
    }

    fn getLayoutConfigKey(tags: u32) u32 {
        switch (context.override_config) {
            .@"per-output" => return 0,
            .@"per-tag" => return tags,
            .@"per-dominant-tag" => return dominantTag(tags),
        }
    }

    fn dominantTag(tags: u32) u5 {
        const bits = @typeInfo(u32).Int.bits;
        var i: u5 = 0;
        while (i < bits) : (i += 1) {
            if ((tags & @shlExact(@as(u32, 1), i)) > 0) return i;
        }
        unreachable;
    }
};

var context = Context{};

pub fn main() !void { // TODO exit value
    defer _ = context.gpa.deinit();
    defer context.config_arena.deinit();

    parseCmdFlags() catch |err| {
        switch (err) {
            error.BadConfig => return,
            error.DumpedInfo => return,
            else => return err,
        }
    };

    parseConfigFile() catch |err| {
        switch (err) {
            error.BadConfig => return,
            error.FileNotFound => {
                log.err("no config file found", .{});
                return;
            },
            else => return err,
        }
    };

    log.debug("connecting to server", .{});
    const display = wl.Display.connect(null) catch {
        log.err("Can not connect to Wayland server.\n", .{});
        return;
    };
    defer display.disconnect();

    const registry = try display.getRegistry();
    registry.setListener(*Context, registryListener, &context); // TODO maybe change zig-wlroots to accept void for user-data pointers
    defer registry.destroy();

    // Will be destroyed in the listener.
    const sync = try display.sync();
    sync.setListener(*Context, syncListener, &context);

    while (context.loop) {
        if (display.dispatch() != .SUCCESS) return error.Whooops;
    }

    // TODO cleanup
}

fn registryListener(registry: *wl.Registry, event: wl.Registry.Event, _: *Context) void {
    switch (event) {
        .global => |ev| {
            if (mem.orderZ(u8, ev.interface, river.LayoutManagerV3.getInterface().name) == .eq) {
                log.debug("binding river_layout_manager_v3", .{});
                context.layout_manager = registry.bind(ev.name, river.LayoutManagerV3, 2) catch |err| {
                    log.err("failed to bind river_layout_manager_v3 : {s}", .{@errorName(err)});
                    return;
                };
            } else if (mem.orderZ(u8, ev.interface, wl.Output.getInterface().name) == .eq) {
                context.newOutput(registry, ev.name) catch |err| {
                    log.err("failed to bind output {}: {s}", .{ ev.name, @errorName(err) });
                    return;
                };
            }
        },
        .global_remove => |ev| {
            var it = context.outputs.first;
            while (it) |node| : (it = node.next) {
                const output = &node.data;
                if (output.global_name == ev.name) {
                    context.outputs.remove(node);
                    output.deinit();
                    const alloc = context.gpa.allocator();
                    alloc.destroy(node);
                    break;
                }
            }
        },
    }
}

fn syncListener(callback: *wl.Callback, _: wl.Callback.Event, _: *Context) void {
    callback.destroy();
    log.debug("sync", .{});

    if (context.layout_manager == null) {
        log.err("server does not support the river_layout_v3 protocol", .{});
        context.loop = false;
        return;
    }

    // There is a chance that an output has been bound and it's name / make_model
    // has been received before the layout_manager has been bound. In that case
    // the output will not have a layout yet.
    var it = context.outputs.first;
    while (it) |node| : (it = node.next) {
        const output = &node.data;
        output.bindLayout() catch |err| {
            log.err("failed to configure output {}: {s}", .{ output.global_name, @errorName(err) });
        };
    }
}

fn parseCmdFlags() !void {
    var i: usize = 1;
    while (i < os.argv.len) : (i += 1) {
        const arg = mem.span(os.argv[i]);
        if (mem.eql(u8, arg, "-h") or mem.eql(u8, arg, "--help")) {
            try dumpUsage();
            return error.DumpedInfo;
        } else if (mem.eql(u8, arg, "-v") or mem.eql(u8, arg, "--version")) {
            log.info("version: {s}", .{build_options.version});
            return error.DumpedInfo;
        } else {
            log.err("unknown flag: {s}", .{arg});
            return error.BadConfig;
        }
    }
}

fn dumpUsage() !void {
    _ = try io.getStdErr().writer().write(
        \\Usage: stacktile [options...]
        \\  -h, --help      Print this help text and exit.
        \\  -v, --version   Print version and exit.
        \\
    );
}

fn parseConfigFile() !void {
    const _alloc = context.gpa.allocator();
    context.config_arena = heap.ArenaAllocator.init(_alloc);

    const alloc = context.config_arena.allocator();
    try context.output_configs.ensureTotalCapacity(alloc, 5);
    try context.layouts.ensureTotalCapacity(alloc, 10);

    const path = blk: {
        if (std.posix.getenv("XDG_CONFIG_HOME")) |xdg_config_home| {
            break :blk try fs.path.join(_alloc, &[_][]const u8{
                xdg_config_home,
                "stacktile/config.ini",
            });
        } else if (std.posix.getenv("HOME")) |home| {
            break :blk try fs.path.join(_alloc, &[_][]const u8{
                home,
                ".config/stacktile/config.ini",
            });
        } else {
            break :blk try _alloc.dupe(u8, "/etc/stacktile/config.ini");
        }
    };
    defer _alloc.free(path);
    try std.posix.access(path, std.posix.R_OK);

    const file = try fs.cwd().openFile(path, .{});
    defer file.close();

    const Current = union(enum) {
        none: void,
        global_config: void,
        layout_config: *Layout,
        output_config: *OutputConfig,
    };
    var current: Current = .none;

    var buffer = io.bufferedReader(file.reader());
    var it = ini.tokenize(buffer.reader());
    var line: usize = 0;
    while (it.next(&line) catch |err| {
        if (err == error.InvalidLine) {
            log.err("{s}:{}: Syntax error.", .{ path, line });
            return error.BadConfig;
        } else {
            return err;
        }
    }) |content| {
        switch (content) {
            .section => |sec| {
                if (mem.eql(u8, sec, "config")) {
                    current = .global_config;
                } else if (mem.startsWith(u8, sec, "layout:")) {
                    current = .{ .layout_config = try context.newLayout(path, line, sec["layout:".len..]) };
                } else if (mem.startsWith(u8, sec, "output:")) {
                    current = .{ .output_config = try context.newOutputConfig(path, line, sec["output:".len..]) };
                } else {
                    log.err("{s}:{}: unknown section: \"{s}\".", .{ path, line, sec });
                    return error.BadConfig;
                }
            },
            .assign => |as| switch (current) {
                .none => {
                    log.err("{s}:{}: assign outside of section", .{ path, line });
                    return error.BadConfig;
                },
                .global_config => try context.handleConfigAssign(path, line, as.variable, as.value),
                .layout_config => try current.layout_config.handleConfigAssign(path, line, as.variable, as.value),
                .output_config => try current.output_config.handleConfigAssign(path, line, as.variable, as.value),
            },
        }
    }

    for (context.output_configs.items) |oc| {
        try oc.getLayouts();
    }
}
